import base64
import os.path
import random
import uuid
from datetime import timedelta
from os import urandom

from botocore.exceptions import ClientError
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import HStoreField
from django.core.validators import MinLengthValidator, MinValueValidator, RegexValidator
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from enumfields import Enum, EnumField

from arkindex.ponos.managers import FarmManager, TaskManager
from arkindex.project.aws import S3FileMixin
from arkindex.project.validators import MaxValueValidator


def generate_seed() -> str:
    return f"{random.getrandbits(256):064x}"


def gen_nonce(size=16):
    """
    Generates a simple nonce
    Number size is defined in bytes (defaults to 128 bits)
    https://cryptography.io/en/latest/glossary/#term-nonce
    """
    return urandom(size)


class Farm(models.Model):
    """
    A group of agents, whose ID and seed can be used to register new agents
    automatically in auto-scaling contexts.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    seed = models.CharField(
        max_length=64,
        default=generate_seed,
        validators=[RegexValidator(r"^[0-9a-f]{64}$")],
    )
    memberships = GenericRelation("users.Right", "content_id")

    objects = FarmManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                "seed",
                name="unique_farm_seed",
            ),
        ]

    def __str__(self) -> str:
        return f"Farm {self.name}"

    def is_available(self, user) -> bool:
        return True


class AgentMode(Enum):
    Docker = "docker"
    Slurm = "slurm"


class Agent(models.Model):
    """
    A remote host that can run tasks.
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    farm = models.ForeignKey(Farm, on_delete=models.PROTECT)
    public_key = models.TextField()
    mode = EnumField(AgentMode, default=AgentMode.Docker, max_length=20)

    hostname = models.SlugField(max_length=64, db_index=False)
    cpu_cores = models.PositiveSmallIntegerField(null=True, blank=True, validators=[MinValueValidator(1)])
    cpu_frequency = models.BigIntegerField(null=True, blank=True, validators=[MinValueValidator(1)])
    # Total amount of RAM on this agent in bytes
    ram_total = models.BigIntegerField(null=True, blank=True, validators=[MinValueValidator(1)])
    # Last minute average CPU load measure on this agent
    cpu_load = models.FloatField(null=True, blank=True)
    # Last RAM load measure expressed as a percentage (0 ≤ ram_load ≤ 1)
    ram_load = models.FloatField(null=True, blank=True)
    last_ping = models.DateTimeField(editable=False)

    class Meta:
        constraints = [
            models.CheckConstraint(
                check=Q(mode=AgentMode.Slurm) | Q(cpu_cores__isnull=False, cpu_frequency__isnull=False, ram_total__isnull=False),
                name="slurm_or_hardware_requirements",
            ),
        ]

    def __str__(self) -> str:
        return self.hostname


class State(Enum):
    """
    Describes the possible :class:`Task` states.

    Inspired by
    `Taskcluster's states <https://docs.taskcluster.net/docs/manual/tasks/runs>`_
    """

    Unscheduled = "unscheduled"
    """
    This is the default initial state of all tasks. Tasks will keep this state in two cases:

    * They do not have dependencies in a :class:`Workflow`, but no :class:`Agent`
      has yet been assigned to them;
    * They depend on other tasks that are not yet :attr:`~State.Completed`.
    """

    Pending = "pending"
    """
    Intermediate state where a task is being assigned to an :class:`Agent` and starting.

    Tasks without dependencies will enter this state when they are assigned to an agent.
    Tasks with dependencies will reach this state as soon as all their dependencies are met.

    Once a task is assigned to an agent, the agent is responsible for updating its state.
    """

    Running = "running"
    """
    State where a task is currently being run by an :class:`Agent`.
    Agents will update the state of their assigned pending tasks to Running as soon
    as the Docker container has started. API users may start monitoring for logs.
    """

    Completed = "completed"
    """
    State where a task has finished successfully.
    Reaching this state means the artifacts are available, and any dependent tasks with
    all their dependencies met will enter the Pending state.
    """

    Failed = "failed"
    """
    State where a task has been run, but ended with a non-zero error code.
    More information might be found in the task's logs.
    """

    Error = "error"
    """
    State where a task may or may not have been run, but an error occurred on the
    :class:`Agent`'s side. This is usually caused by Docker or host-related issues.
    """

    Stopping = "stopping"
    """
    State that can be set by API users to ask for a running task to be stopped.
    The task's assigned agent will be instructed to stop the task, and will report the
    Stopped state when done.
    """

    Stopped = "stopped"
    """
    State where a task that entered the Stopping state has successfully stopped.
    """


# States where a task is considered final.
# Once a task reaches a final state, its state should no longer change.
FINAL_STATES = (
    State.Completed,
    State.Failed,
    State.Error,
    State.Stopped,
)

# States where a task could start, be running, or still be using some resources, on an agent
ACTIVE_STATES = (
    State.Unscheduled,
    State.Pending,
    State.Running,
    State.Stopping,
)

# Tasks priority to determine the overall state of multiple tasks
# If there are failed tasks, the process is failed.
# Else, if there are any errors, the process errored.
# If there are any running tasks, no matter in which state another task may be,
# the process is set as running to prevent retrying and allow stopping.
STATES_ORDERING = [
    State.Running,
    State.Failed,
    State.Error,
    State.Stopping,
    State.Stopped,
    State.Pending,
    State.Unscheduled,
    State.Completed,
]


def expiry_default():
    """
    Default value for Task.expiry.

    :rtype: datetime
    """
    return timezone.now() + timedelta(days=settings.PONOS_TASK_EXPIRY)


def task_token_default():
    """
    Default value for Task.token.

    :rtype: str
    """
    return base64.encodebytes(uuid.uuid4().bytes + uuid.uuid4().bytes).strip().decode("utf-8")


class TaskLogs(S3FileMixin):
    s3_bucket = settings.PONOS_S3_LOGS_BUCKET

    def __init__(self, task) -> None:
        self.task = task

    @property
    def s3_key(self) -> str:
        return f"{self.task.id}.log"

    @property
    def latest(self) -> str:
        """
        Fetch the last N bytes from the task's log file on S3, and return them decoded.
        An empty string is returned if the logs file is missing.
        The decoded text is cut to start after the first found newline character to avoid incomplete lines.

        The logs will start with ``[Logs were truncated]`` to help API users or frontends
        make it clear there are more logs available, unless less bytes than the requested amount are returned.

        The amount of fetched bytes is defined by the ``PONOS_LOG_TAIL`` setting, which defaults to 10000.
        """
        max_length = getattr(settings, "PONOS_LOG_TAIL", 10000)

        try:
            log_bytes = self.s3_object.get(
                Range=f"bytes=-{max_length}",
            )["Body"].read()

        except ClientError as e:
            # When the log file does not exist, HeadObject returns a 404 code and GetObject returns NoSuchKey,
            # because HEAD requests do not have response bodies, so S3 only returns a status code
            # with no error code. We are using GetObject here, so we will get NoSuchKey.
            # When the log file exists but has a length of zero bytes, and we use a `Range`,
            # then Ceph returns InvalidRange while MinIO and AWS will return a successful response with zero bytes.
            if e.response["Error"]["Code"] in ("NoSuchKey", "InvalidRange"):
                return ""
            raise

        text = log_bytes.decode("utf-8", errors="replace")

        if len(text) < max_length:
            return text

        # Not starting from the beginning of the file
        # Start at the next new line
        text = text[text.find("\n") + 1 :]
        return "[Logs were truncated]\n" + text


class Task(models.Model):

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    run = models.PositiveIntegerField()
    depth = models.PositiveIntegerField()
    # blank=False is only used in Django form validation, to prevent blank slugs we use a minimum length
    slug = models.CharField(max_length=250, validators=[MinLengthValidator(1)])
    priority = models.PositiveIntegerField(default=10)
    state = EnumField(State, default=State.Unscheduled, max_length=20)
    image = models.CharField(max_length=250)
    shm_size = models.CharField(max_length=80, blank=True, null=True, editable=False)
    command = models.TextField(blank=True, null=True)
    env = HStoreField(default=dict)

    agent = models.ForeignKey(
        Agent,
        related_name="tasks",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    requires_gpu = models.BooleanField(default=False)
    gpu = models.ForeignKey(
        "ponos.GPU",
        related_name="tasks",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    process = models.ForeignKey(
        "process.Process",
        related_name="tasks",
        on_delete=models.CASCADE,
    )
    worker_run = models.ForeignKey(
        "process.WorkerRun",
        on_delete=models.SET_NULL,
        related_name="tasks",
        null=True,
        blank=True,
    )
    parents = models.ManyToManyField(
        "self",
        related_name="children",
        symmetrical=False,
    )
    container = models.CharField(
        max_length=64,
        null=True,
        blank=True,
    )
    original_task = models.ForeignKey(
        "self",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="restarts"
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    expiry = models.DateTimeField(default=expiry_default)

    # Remote files required to start the container
    extra_files = HStoreField(default=dict, blank=True)

    token = models.CharField(
        default=task_token_default,
        # The token generation always returns 52 characters
        max_length=52,
    )

    objects = TaskManager()

    class Meta:
        ordering = ("process_id", "run", "depth", "slug")
        constraints = [
            models.UniqueConstraint(
                "process",
                "run",
                "slug",
                name="unique_process_run_slug",
            ),
            models.UniqueConstraint(
                "gpu",
                condition=Q(state__in=ACTIVE_STATES),
                name="unique_gpu_on_active_tasks",
            ),
            models.UniqueConstraint(
                "token",
                name="unique_task_token",
            ),
        ]

    def __str__(self) -> str:
        return f"Task {self.slug}, run {self.run}, depth {self.depth}"

    def get_absolute_url(self) -> str:
        """
        :returns: URL to the task details API for this task.
        """
        return reverse("api:task-details", args=[self.id])

    def is_final(self) -> bool:
        """
        Helper to tell whether a task is final.
        A task is considered final when it is in one of the final states:
        :attr:`~State.Completed`, :attr:`~State.Failed`, :attr:`~State.Error`, :attr:`~State.Stopped`

        :returns: Whether or not the task is considered final.
        """
        return self.state in FINAL_STATES

    @property
    def logs(self) -> TaskLogs:
        return TaskLogs(self)


def artifact_max_size():
    """
    AWS restricts uploads to 5GiB per PUT request,
    but some S3 implementations might not have this restriction,
    so we default to 5 GiB and allow overriding through the Django settings.

    We use a function to allow this setting to change at runtime,
    which makes unit testing a lot easier.
    """
    setting = getattr(settings, "PONOS_ARTIFACT_MAX_SIZE", None)
    return setting if setting is not None else 5 * 1024**3


class Artifact(S3FileMixin, models.Model):
    """
    A task Artifact (Json report, docker images, ML Models...)
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    task = models.ForeignKey(Task, related_name="artifacts", on_delete=models.CASCADE)
    path = models.CharField(max_length=500)
    size = models.BigIntegerField(
        validators=[
            MinValueValidator(1),
            MaxValueValidator(artifact_max_size),
        ]
    )
    content_type = models.CharField(max_length=250, default="application/octet-stream")
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ("task", "path")
        unique_together = (("task", "path"),)

    s3_bucket = settings.PONOS_S3_ARTIFACTS_BUCKET

    @property
    def s3_key(self):
        return os.path.join(
            str(self.task.id),
            str(self.path),
        )


class Secret(models.Model):
    """
    A secret encrypted with a derivate of the Ponos server private key (ECDH)
    Secret content is encrypted with AES, using a key generated serializing ponos private key (ECDH) with a salt
    """

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=250)
    nonce = models.BinaryField(max_length=16, default=gen_nonce)
    content = models.BinaryField(editable=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                "name",
                name="unique_secret_names",
            ),
        ]

    def __str__(self):
        return self.name


class GPU(models.Model):
    """
    A Graphic Card, attached to an agent
    """

    # The specific GPU uuid defined by the card itself
    # It's used as GPU-{id} to identity the GPU for docker
    id = models.UUIDField(primary_key=True)
    agent = models.ForeignKey(Agent, related_name="gpus", on_delete=models.CASCADE)
    name = models.CharField(max_length=250)

    # The numeric index assigned to the card by the host
    index = models.PositiveIntegerField()

    # RAM management
    ram_total = models.BigIntegerField(validators=[MinValueValidator(1)])

    class Meta:
        unique_together = (("agent_id", "index"),)

    def __str__(self):
        return self.name
