import uuid
from itertools import combinations
from unittest.mock import MagicMock, call, patch

from django.urls import reverse
from django.utils import timezone as DjangoTimeZone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessDatasetSet, ProcessMode
from arkindex.project.tests import FixtureAPITestCase, force_constraints_immediate
from arkindex.project.tools import fake_now
from arkindex.training.models import Dataset, DatasetElement, DatasetSet, DatasetState
from arkindex.users.models import Role, User

# Using the fake DB fixtures creation date when needed
FAKE_CREATED = "2020-02-02T01:23:45.678000Z"


class TestDatasetsAPI(FixtureAPITestCase):
    @classmethod
    @patch.object(DjangoTimeZone, "now", fake_now)
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        cls.task = cls.process.tasks.first()
        cls.dataset_creator = User.objects.create(email="adam@nerv.co.jp", display_name="Adam", verified_email=True)
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.read_user = User.objects.get(email="user3@user.fr")
        cls.write_user = User.objects.get(email="user2@user.fr")
        cls.dataset = Dataset.objects.get(name="First Dataset")
        cls.dataset2 = Dataset.objects.get(name="Second Dataset")
        ProcessDatasetSet.objects.bulk_create(
            ProcessDatasetSet(process=cls.process, set=ds)
            for ds in cls.dataset.sets.all()
        )
        ProcessDatasetSet.objects.create(process=cls.process, set=cls.dataset2.sets.get(name="test"))
        cls.private_dataset = Dataset.objects.create(name="Private Dataset", description="Dead Sea Scrolls", corpus=cls.private_corpus, creator=cls.dataset_creator)
        cls.private_dataset_set = DatasetSet.objects.create(dataset=cls.private_dataset, name="Private set")
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.page1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.page2 = cls.corpus.elements.get(name="Volume 1, page 1v")
        cls.page3 = cls.corpus.elements.get(name="Volume 1, page 2r")

    # ListCorpusDatasets

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_list_private_corpus(self, has_access_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.private_corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this corpus."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.private_corpus, Role.Guest.value, skip_public=False))

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["results"], [
            {
                "name": "First Dataset",
                "description": "dataset number one",
                "id": str(self.dataset.pk),
                "sets": [
                    {
                        "id": str(ds.id),
                        "name": ds.name
                    }
                    for ds in self.dataset.sets.order_by("name")
                ],
                "set_elements": None,
                "state": "open",
                "creator": "Test user",
                "task_id": None,
                "corpus_id": str(self.corpus.id),
                "unique_elements": True,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED,
            },
            {
                "name": "Second Dataset",
                "description": "dataset number two",
                "id": str(self.dataset2.pk),
                "sets": [
                    {
                        "id": str(ds.id),
                        "name": ds.name
                    }
                    for ds in self.dataset2.sets.order_by("name")
                ],
                "set_elements": None,
                "state": "open",
                "creator": "Test user",
                "task_id": None,
                "corpus_id": str(self.corpus.id),
                "unique_elements": True,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED,
            }
        ])

    # CreateDataset

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_private_corpus(self, has_access_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.private_corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to this corpus."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.private_corpus, Role.Contributor.value, skip_public=False))

    def test_create_name_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field is required."]})

    def test_create_name_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_name_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "        ", "description": "My dataset for my experiments."},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                    "description": "My dataset for my experiments."
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_create_name_already_exists_in_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "First Dataset",
                    "description": "My dataset for my experiments."
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_create_description_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field is required."]})

    def test_create_description_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": ""},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field may not be blank."]})

    def test_create_description_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "             "},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field may not be blank."]})

    def test_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "unique_elements": False},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": [
                {
                    "id": str(ds.id),
                    "name": ds.name
                }
                for ds in created_dataset.sets.order_by("name")
            ],
            "set_elements": {
                "train": 0,
                "test": 0,
                "dev": 0,
            },
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "unique_elements": False,
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_state_ignored(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "state": "complete"},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertEqual(created_dataset.state, DatasetState.Open)
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": [
                {
                    "id": str(ds.id),
                    "name": ds.name
                }
                for ds in created_dataset.sets.order_by("name")
            ],
            "set_elements": {
                "train": 0,
                "test": 0,
                "dev": 0,
            },
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "unique_elements": True,
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_sets(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "set_names": ["a", "b", "c", "d"]},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": [
                {
                    "id": str(ds.id),
                    "name": ds.name
                }
                for ds in created_dataset.sets.order_by("name")
            ],
            "set_elements": {"a": 0, "b": 0, "c": 0, "d": 0},
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "unique_elements": True,
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_sets_empty_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "set_names": []},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set_names": ["Either do not specify set names to use the default values, or specify a non-empty list of names."]})

    def test_create_sets_unique_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "set_names": ["a", "a", "b"]},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set_names": ["Set names must be unique."]})

    def test_create_sets_blank_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "set_names": ["     ", " ", "b"]},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_names":
                {
                    "0": ["This field may not be blank."],
                    "1": ["This field may not be blank."]
                }
        })

    def test_create_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "My Dataset",
                    "description": "My dataset for my experiments.",
                    "set_names": ["unit-00", "Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor."]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_names": {
                "1": ["Ensure this field has no more than 50 characters."]
            }
        })

    # UpdateDataset

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_update_requires_write_corpus(self, has_access_mock):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.read_user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_update_doesnt_exist(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Dataset matches the given query."})

    def test_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                    "description": "My dataset for my experiments.",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Another Dataset",
                    "description": "My dataset for my experiments.",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_update_requires_all_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field is required."]})

    def test_update_sets_forbidden(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.dataset.task_id)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "set_names": ["unit-01", "unit-02"],
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set_names": ["This API endpoint does not allow updating a dataset's sets."]})

    def test_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "",
                    "description": "          ",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["This field may not be blank."],
            "description": ["This field may not be blank."]
        })

    def test_update_all_errors(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "",
                    "description": "          ",
                    "set_names": ["unit-00", "unit-00", "unit-02"]
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["This field may not be blank."],
            "description": ["This field may not be blank."],
            "set_names": ["This API endpoint does not allow updating a dataset's sets."]
        })

    def test_update(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "a", "description": "a", "unique_elements": False},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.name, "a")
        self.assertEqual(self.dataset.description, "a")
        self.assertEqual(self.dataset.unique_elements, False)

    def test_update_forbidden_unique_elements(self):
        self.dataset.unique_elements = False
        self.dataset.save()
        _, test_set, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.vol)
        test_set.set_elements.create(element=self.vol)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "a", "description": "a", "unique_elements": True},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "unique_elements": ["Some elements are currently contained by multiple sets."]
        })

    def test_update_state_complete_removes_artifact(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.task = self.task
        self.dataset.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "open"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(response.json()["set_elements"], {"test": 0, "train": 0, "dev": 0})
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.task, None)
        self.assertEqual(self.dataset.task_id, None)

    def test_update_complete_ponos_task(self):
        """It is possible to use Ponos task auth to update Dataset's state to complete"""
        self.dataset.state = DatasetState.Building
        self.dataset.save()
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "complete"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(response.json()["set_elements"], {"test": 0, "train": 0, "dev": 0})
        self.assertEqual(self.dataset.state, DatasetState.Complete)
        self.assertEqual(self.dataset.task_id, self.task.id)

    def test_update_task_state_forbidden_ponos_auth(self):
        """Dataset's state update is limited to specific transitions with Ponos authentication"""
        op, build, complete, error = (DatasetState[state] for state in ("Open", "Building", "Complete", "Error"))
        states = {
            (op, op): True,
            (op, build) : True,
            (op, complete) : False,
            (op, error): False,
            (build, op): False,
            (build, build) : True,
            (build, complete) : True,
            (build, error): True,
            (error, op): False,
            (error, build) : True,
            (error, complete) : False,
            (error, error): True,
            (complete, op): False,
            (complete, build): False,
            (complete, complete): True,
            (complete, error): False,
        }
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.put(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                    data={
                        "name": self.dataset.name,
                        "description": self.dataset.description,
                        "state": state_to.value,
                    },
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_update_task_state_forbidden_user_auth(self):
        """Dataset's state update is limited to specific transitions without Ponos authentication"""
        states = {couple: False for couple in combinations(DatasetState, 2)}
        states[(DatasetState.Error, DatasetState.Open)] = True
        states[(DatasetState.Complete, DatasetState.Open)] = True
        self.client.force_login(self.user)
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.put(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    data={
                        "name": self.dataset.name,
                        "description": self.dataset.description,
                        "state": state_to.value,
                    },
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_update_complete_update_attrs(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "AA",
                    "description": "BB",
                    "unique_elements": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.name, "AA")
        self.assertEqual(self.dataset.description, "BB")
        self.assertEqual(self.dataset.unique_elements, False)

    def test_update_state_requires_dataset_in_process(self):
        self.process.process_sets.all().delete()
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "building"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": ["A task can only update the state of one of the datasets of its process."]
        })

    def test_update_ponos_task_bad_state(self):
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "decommissioned"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ["Value is not of type DatasetState"]})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    # PartialUpdateDataset

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_partial_update_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_partial_update_requires_write_corpus(self, has_access_mock):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.read_user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_partial_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_partial_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Another Dataset",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_partial_update(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"description": "Omedeto!", "unique_elements": False},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, "First Dataset")
        self.assertEqual(self.dataset.description, "Omedeto!")
        self.assertEqual(self.dataset.unique_elements, False)
        self.assertCountEqual(list(self.dataset.sets.values_list("name", flat=True)), ["train", "test", "dev"])

    def test_partial_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "                  ",
                    "description": "",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."], "description": ["This field may not be blank."]})

    def test_partial_update_forbidden_unique_elements(self):
        self.dataset.unique_elements = False
        self.dataset.save()
        _, test_set, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.vol)
        test_set.set_elements.create(element=self.vol)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "a", "description": "a", "unique_elements": True},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "unique_elements": ["Some elements are currently contained by multiple sets."]
        })

    def test_partial_update_state_open_removes_artifacts(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.task = self.task
        self.dataset.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"state": "open"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(response.json()["set_elements"], {"test": 0, "train": 0, "dev": 0})
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.task, None)
        self.assertEqual(self.dataset.task_id, None)

    def test_partial_update_state_complete_ponos_task(self):
        self.dataset.state = DatasetState.Building
        self.dataset.save()
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={"state": "complete"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(response.json()["set_elements"], {"test": 0, "train": 0, "dev": 0})
        self.assertEqual(self.dataset.state, DatasetState.Complete)
        self.assertEqual(self.dataset.task_id, self.task.id)

    def test_partial_update_state_requires_dataset_in_process(self):
        self.process.process_sets.all().delete()
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={"state": "building"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": ["A task can only update the state of one of the datasets of its process."]
        })

    def test_partial_update_ponos_task_bad_state(self):
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "state": "decommissioned"
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ["Value is not of type DatasetState"]})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    def test_partial_update_ponos_task_state_forbidden(self):
        """Dataset's state update is limited to specific transitions"""
        op, build, complete, error = (DatasetState[state] for state in ("Open", "Building", "Complete", "Error"))
        states = {
            (op, op): True,
            (op, build) : True,
            (op, complete) : False,
            (op, error): False,
            (build, op): False,
            (build, build) : True,
            (build, complete) : True,
            (build, error): True,
            (error, op): False,
            (error, build) : True,
            (error, complete) : False,
            (error, error): True,
            (complete, op): False,
            (complete, build): False,
            (complete, complete): True,
            (complete, error): False,
        }
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.patch(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                    data={
                        "name": self.dataset.name,
                        "description": self.dataset.description,
                        "state": state_to.value,
                    },
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_partial_update_task_state_forbidden_user_auth(self):
        """Dataset's state update is limited to specific transitions without Ponos authentication"""
        states = {couple: False for couple in combinations(DatasetState, 2)}
        states[(DatasetState.Error, DatasetState.Open)] = True
        states[(DatasetState.Complete, DatasetState.Open)] = True
        self.client.force_login(self.user)
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.patch(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    data={"state": state_to.value},
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_partial_update_complete_update_attrs(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "AA",
                    "description": "BB",
                    "unique_elements": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.name, "AA")
        self.assertEqual(self.dataset.description, "BB")
        self.assertEqual(self.dataset.unique_elements, False)

    # RetrieveDataset

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_retrieve_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_retrieve_requires_read_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.private_dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Dataset matches the given query."})

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_retrieve_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Dataset matches the given query."})

    def test_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.dataset.id),
            "name": "First Dataset",
            "description": "dataset number one",
            "state": "open",
            "sets": [
                {
                    "id": str(ds.id),
                    "name": ds.name
                }
                for ds in self.dataset.sets.order_by("name")
            ],
            "set_elements": {"test": 0, "train": 0, "dev": 0},
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "unique_elements": True,
            "created": FAKE_CREATED,
            "updated": FAKE_CREATED
        })

    def test_retrieve_task_id(self):
        self.client.force_login(self.user)
        self.dataset.task = self.task
        self.dataset.save()
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["task_id"], str(self.task.id))

    # DestroyDataset

    def test_delete_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_delete_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_delete_requires_corpus_admin(self, has_access_mock):
        self.client.force_login(self.write_user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to corpus Unit Tests."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.write_user, self.corpus, Role.Admin.value, skip_public=False))

    def test_delete_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No Dataset matches the given query."})

    def test_delete_dataset_in_process_forbidden(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ["This dataset cannot be deleted because at least one of its sets is used in a process."])
        self.dataset.refresh_from_db()

    def test_delete(self):
        self.client.force_login(self.user)
        # Remove dataset sets from process
        ProcessDatasetSet.objects.filter(process_id=self.process.id, set__dataset_id=self.dataset.id).delete()
        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Dataset.DoesNotExist):
            self.dataset.refresh_from_db()

    @patch("arkindex.project.mixins.has_access", return_value=True)
    def test_delete_not_open(self, has_access_mock):
        self.client.force_login(self.write_user)
        cases = [DatasetState.Building, DatasetState.Complete, DatasetState.Error]
        for state in cases:
            with self.subTest(state=state):
                self.dataset.state = state
                self.dataset.save()
                with self.assertNumQueries(3):
                    response = self.client.delete(
                        reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), ["Only datasets in an open state can be deleted."])
            self.assertEqual(has_access_mock.call_count, 1)
            self.assertEqual(has_access_mock.call_args, call(self.write_user, self.corpus, Role.Admin.value, skip_public=False))
            has_access_mock.reset_mock()

    def test_delete_elements(self):
        """
        DestroyDataset also deletes DatasetElements
        """
        dev_set, test_set, train_set = self.dataset.sets.all().order_by("name")
        test_set.set_elements.create(element_id=self.vol.id)
        train_set.set_elements.create(element_id=self.vol.id, set="train")
        train_set.set_elements.create(element_id=self.page1.id, set="train")
        dev_set.set_elements.create(element_id=self.page2.id, set="dev")
        dev_set.set_elements.create(element_id=self.page3.id, set="dev")
        # Remove dataset sets from process
        ProcessDatasetSet.objects.filter(process_id=self.process.id, set__dataset_id=self.dataset.id).delete()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Dataset.DoesNotExist):
            self.dataset.refresh_from_db()

        self.assertFalse(DatasetElement.objects.filter(set__dataset_id=self.dataset.id).exists())

        # No elements should have been deleted
        self.vol.refresh_from_db()
        self.page1.refresh_from_db()
        self.page2.refresh_from_db()
        self.page3.refresh_from_db()

    def test_list_elements_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(self.dataset.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_elements_invalid_dataset_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(uuid.uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_list_elements_readable_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(self.private_dataset.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_list_elements_set_filter_wrong_set(self):
        test_set = self.dataset.sets.get(name="test")
        test_set.set_elements.create(element_id=self.page1.id)
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:dataset-elements", kwargs={"pk": str(self.dataset.id)}),
                data={"set": "train"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": None,
            "next": None,
            "previous": None,
            "results": [],
        })

    def test_list_elements_set_filter(self):
        _, test_set, train_set = self.dataset.sets.all().order_by("name")
        test_set.set_elements.create(element_id=self.page1.id)
        train_set.set_elements.create(element_id=self.page2.id)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.pk}),
                data={"set": "train", "with_count": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data.pop("results")
        self.assertDictEqual(data, {"count": 1, "next": None, "previous": None})
        self.assertEqual(len(results), 1)
        dataset_element = results[0]
        self.assertEqual(dataset_element["element"]["id"], str(self.page2.id))
        self.assertEqual(dataset_element["set"], "train")

    @patch("arkindex.documents.models.Element.thumbnail", MagicMock(s3_url="s3_url"))
    def test_list_elements(self):
        dev_set, test_set, train_set = self.dataset.sets.all().order_by("name")
        test_set.set_elements.create(element_id=self.vol.id)
        train_set.set_elements.create(element_id=self.page1.id)
        dev_set.set_elements.create(element_id=self.page2.id)
        dev_set.set_elements.create(element_id=self.page3.id)
        self.page3.confidence = 0.42
        self.page3.mirrored = True
        self.page3.rotation_angle = 42
        self.page3.save()

        # The elements retrieved when listing dataset elements are ordered by element id so this
        # test needs to be insensitive to element id changes when regenerating test fixtures
        elements = {
            self.page1.id: {
                "set": "train",
                "element": {
                    "id": str(self.page1.id),
                    "confidence": None,
                    "type": "page",
                    "mirrored": False,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page1.name,
                    "zone": {
                        "id": str(self.page1.id),
                        "image": {
                            "id": str(self.page1.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img1",
                            "path": "img1",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img1/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 0,
                    "classifications": None,
                    "has_children": None,
                    "metadata": None,
                    "transcriptions": None,
                    "worker_run": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.page2.id: {
                "set": "dev",
                "element": {
                    "id": str(self.page2.id),
                    "confidence": None,
                    "type": "page",
                    "mirrored": False,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page2.name,
                    "zone": {
                        "id": str(self.page2.id),
                        "image": {
                            "id": str(self.page2.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img2",
                            "path": "img2",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img2/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 0,
                    "classifications": None,
                    "has_children": None,
                    "metadata": None,
                    "transcriptions": None,
                    "worker_run": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.page3.id: {
                "set": "dev",
                "element": {
                    "id": str(self.page3.id),
                    "confidence": 0.42,
                    "type": "page",
                    "mirrored": True,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page3.name,
                    "zone": {
                        "id": str(self.page3.id),
                        "image": {
                            "id": str(self.page3.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img3",
                            "path": "img3",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img3/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 42,
                    "classifications": None,
                    "has_children": None,
                    "metadata": None,
                    "transcriptions": None,
                    "worker_run": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.vol.id: {
                "set": "test",
                "element": {
                    "id": str(self.vol.id),
                    "confidence": None,
                    "type": "volume",
                    "mirrored": False,
                    "name": "Volume 1",
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "zone": None,
                    "rotation_angle": 0,
                    "classifications": None,
                    "has_children": None,
                    "metadata": None,
                    "transcriptions": None,
                    "worker_run": None,
                    "thumbnail_url": "s3_url",
                    "created": FAKE_CREATED
                },
            }
        }

        first_three = [value for key, value in sorted(elements.items())[:3]]
        self.client.force_login(self.read_user)
        for state in DatasetState:
            # A dataset should be readable in any state
            self.dataset.state = state
            self.dataset.save()
            with self.subTest(state=state):
                with self.assertNumQueries(5):
                    response = self.client.get(
                        reverse("api:dataset-elements", kwargs={"pk": self.dataset.pk}),
                        {"page_size": 3},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                data = response.json()
                self.assertTrue("?cursor=" in data["next"])
                self.assertIsNone(data["count"])
                self.assertListEqual(data["results"], first_three)

    def test_add_element_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_add_element_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_add_element_private_dataset(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_add_element_requires_writable_corpus(self, has_access_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            "detail": "You do not have contributor access to the corpus of this dataset."
        })

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_add_element_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_id": ["This field is required."],
            "set": ["This field is required."],
        })

    def test_add_element_wrong_element(self):
        element = self.private_corpus.elements.create(type=self.private_corpus.types.create(slug="folder"))
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(element.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_id": [f'Invalid pk "{element.id}" - object does not exist.'],
        })

    def test_add_element_wrong_set(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "aaaaaaaaaaa", "element_id": str(self.vol.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set": ["Object with name=aaaaaaaaaaa does not exist."],
        })

    def test_add_element_dataset_requires_open(self):
        self.client.force_login(self.user)
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.vol.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["You can only add elements to a dataset in an open state."])

    def test_add_element_already_exists(self):
        test_set = self.dataset.sets.get(name="test")
        test_set.set_elements.create(element=self.page1)
        self.dataset.unique_elements = False
        self.dataset.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.page1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This element is already part of this set."]})

    def test_add_element(self):
        train_set = self.dataset.sets.get(name="train")
        self.client.force_login(self.user)
        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "train", "element_id": str(self.page1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertQuerySetEqual(
            train_set.set_elements.values_list("set__name", "element__name").order_by("element__name"),
            [("train", "Volume 1, page 1r")]
        )

    def test_add_element_forbid_dupes(self):
        train_set = self.dataset.sets.get(name="train")
        train_set.set_elements.create(element=self.page1)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.page1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_id": ["The dataset requires unique elements and this element is already present in set train."]
        })

    def test_add_element_allow_dupes(self):
        self.dataset.unique_elements = False
        self.dataset.save()
        train_set = self.dataset.sets.get(name="train")
        train_set.set_elements.create(element=self.page1)
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.page1.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertQuerySetEqual(
            (
                DatasetElement.objects
                .filter(set__dataset=self.dataset)
                .values_list("set__name", "element__name")
                .order_by("set__name", "element__name")
            ),
            [
                ("test", "Volume 1, page 1r"),
                ("train", "Volume 1, page 1r"),
            ]
        )

    # CreateDatasetElementSelection

    def test_add_from_selection_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_add_from_selection_forbidden_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("get", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                response = getattr(self.client, method)(
                    reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id})
                )
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_add_from_selection_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_add_from_selection_private_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.private_corpus.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_add_from_selection_requires_writable_corpus(self, has_access_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            "detail": "You need a Contributor access to the corpus to perform this action."
        })

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_add_from_selection_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_id": ["This field is required."],
        })

    def test_add_from_selection_wrong_values(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": "AAA"},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_id": ["“AAA” is not a valid UUID."],
        })

    def test_add_from_selection_wrong_dataset(self):
        self.private_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": self.private_dataset_set.id},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_id": [f"Dataset {self.private_dataset.id} is not part of corpus Unit Tests."],
        })

    def test_add_from_selection_completed_dataset(self):
        """A dataset in the Complete state is immutable"""
        test_set = self.dataset.sets.get(name="test")
        self.client.force_login(self.user)
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": test_set.id},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_id": [f"Dataset {self.dataset.id} is marked as completed."]
        })

    def test_add_from_selection(self):
        train_set = self.dataset.sets.get(name="train")
        train_set.set_elements.create(element=self.page1)
        self.assertQuerySetEqual(
            train_set.set_elements.values_list("set__name", "element__name").order_by("element__name"),
            [("train", "Volume 1, page 1r")]
        )
        self.user.selected_elements.set([self.vol, self.page1, self.page2])

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": str(train_set.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertQuerySetEqual(
            train_set.set_elements.values_list("set__name", "element__name").order_by("element__name"),
            [
                ("train", "Volume 1"),
                ("train", "Volume 1, page 1r"),
                ("train", "Volume 1, page 1v"),
            ]
        )

    def test_add_from_selection_forbid_dupes(self):
        _, test_set, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.vol)
        self.user.selected_elements.set([self.vol, self.page1])

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": str(test_set.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set_id": ["The dataset requires unique elements and some elements are already present in set train."]
        })

    def test_add_from_selection_allow_dupes(self):
        _, test_set, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.vol)
        self.user.selected_elements.set([self.vol, self.page1])
        test_set = self.dataset.sets.get(name="test")
        self.dataset.unique_elements = False
        self.dataset.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set_id": str(test_set.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertQuerySetEqual(
            (
                DatasetElement.objects
                .filter(set__dataset=self.dataset)
                .values_list("set__name", "element__name")
                .order_by("set__name", "element__name")
            ),
            [
                ("test", "Volume 1"),
                ("test", "Volume 1, page 1r"),
                ("train", "Volume 1"),
            ]
        )

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_element_datasets_requires_read_access(self, filter_rights_mock):
        self.client.force_login(self.user)
        private_elt = self.private_corpus.elements.create(type=self.private_corpus.types.create(slug="t"), name="elt")
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:element-sets", kwargs={"pk": private_elt.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    def test_element_datasets_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("post", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                client_method = getattr(self.client, method)
                response = client_method(reverse("api:element-sets", kwargs={"pk": str(self.vol.id)}))
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_element_datasets_public(self):
        """
        A non authenticated user can list datasets of a public element
        """
        train_set = self.dataset.sets.get(name="train")
        train_set.set_elements.create(element=self.vol)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:element-sets", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": None,
                "next": None
            }]
        })

    def test_element_datasets(self):
        self.client.force_login(self.user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.page1, set="train")
        dev_set.set_elements.create(element=self.page1, set="dev")
        train_set_2 = self.dataset2.sets.get(name="train")
        train_set_2.set_elements.create(element=self.page1, set="train")
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:element-sets", kwargs={"pk": str(self.page1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 3,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "dev",
                "previous": None,
                "next": None
            }, {
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": None,
                "next": None
            }, {
                "dataset": {
                    "id": str(self.dataset2.id),
                    "name": "Second Dataset",
                    "description": "dataset number two",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset2.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset2.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset2.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": None,
                "next": None
            }]
        })

    def test_element_datasets_with_neighbors_false(self):
        self.client.force_login(self.user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set_2 = self.dataset2.sets.get(name="train")
        train_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        train_set_2.set_elements.create(element=self.page1)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:element-sets", kwargs={"pk": str(self.page1.id)}), {"with_neighbors": False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 3,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "dev",
                "previous": None,
                "next": None
            }, {
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": None,
                "next": None
            }, {
                "dataset": {
                    "id": str(self.dataset2.id),
                    "name": "Second Dataset",
                    "description": "dataset number two",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset2.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset2.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset2.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": None,
                "next": None
            }]
        })

    def test_element_datasets_with_neighbors(self):
        self.client.force_login(self.user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set_2 = self.dataset2.sets.get(name="train")
        train_set.set_elements.create(element=self.page1)
        train_set.set_elements.create(element=self.page2)
        train_set.set_elements.create(element=self.page3)
        dev_set.set_elements.create(element=self.page1)
        train_set_2.set_elements.create(element=self.page1)
        train_set_2.set_elements.create(element=self.page3)

        # Results are alphabetically ordered and must not depend on the random page UUIDs
        sorted_dataset_elements = sorted([str(self.page1.id), str(self.page2.id), str(self.page3.id)])
        page1_index_1 = sorted_dataset_elements.index(str(self.page1.id))
        sorted_dataset2_elements = sorted([str(self.page1.id), str(self.page3.id)])
        page1_index_2 = sorted_dataset2_elements.index(str(self.page1.id))

        with self.assertNumQueries(8):
            response = self.client.get(reverse("api:element-sets", kwargs={"pk": str(self.page1.id)}), {"with_neighbors": True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 3,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "dev",
                "previous": None,
                "next": None
            }, {
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": (
                    sorted_dataset_elements[page1_index_1 - 1]
                    if page1_index_1 - 1 >= 0
                    else None
                ),
                "next": (
                    sorted_dataset_elements[page1_index_1 + 1]
                    if page1_index_1 + 1 <= 2
                    else None
                )
            }, {
                "dataset": {
                    "id": str(self.dataset2.id),
                    "name": "Second Dataset",
                    "description": "dataset number two",
                    "sets": [
                        {
                            "id": str(ds.id),
                            "name": ds.name
                        }
                        for ds in self.dataset2.sets.order_by("name")
                    ],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "unique_elements": True,
                    "created": self.dataset2.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset2.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
                "previous": (
                    sorted_dataset2_elements[page1_index_2 - 1]
                    if page1_index_2 == 1
                    else None
                ),
                "next": (
                    sorted_dataset2_elements[page1_index_2 + 1]
                    if page1_index_2 == 0
                    else None
                )
            }]
        })

    # DatasetClone

    def test_clone_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_clone_forbidden_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("get", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                response = getattr(self.client, method)(
                    reverse("api:dataset-clone", kwargs={"pk": self.dataset.id})
                )
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_clone_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_clone_private_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_clone_requires_writable_corpus(self, has_access_mock):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            "detail": "You need a Contributor access to the dataset's corpus."
        })

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_clone(self):
        self.dataset.creator = self.superuser
        self.dataset.state = DatasetState.Error
        self.dataset.unique_elements = False
        self.dataset.task = self.task
        self.dataset.save()
        dev_set, test_set, _ = self.dataset.sets.all().order_by("name")
        test_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.vol)
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), ["First Dataset", "Second Dataset"])

        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), [
            "First Dataset",
            "Clone of First Dataset",
            "Second Dataset",
        ])
        data = response.json()
        clone = self.corpus.datasets.get(id=data["id"])
        dev_clone, test_clone, train_clone = clone.sets.all().order_by("name")
        self.assertEqual(clone.creator, self.user)
        self.assertEqual(clone.unique_elements, False)
        data.pop("created")
        data.pop("updated")
        cloned_sets = data.pop("sets")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(clone.id),
                "name": "Clone of First Dataset",
                "description": self.dataset.description,
                "creator": self.user.display_name,
                "corpus_id": str(self.corpus.id),
                "set_elements": {"test": 1, "train": 0, "dev": 2},
                "state": DatasetState.Open.value,
                "task_id": str(self.task.id),
                "unique_elements": False,
            },
        )
        self.assertCountEqual(cloned_sets, [
            {
                "name": "train",
                "id": str(train_clone.id)
            },
            {
                "name": "test",
                "id": str(test_clone.id)
            },
            {
                "name": "dev",
                "id": str(dev_clone.id)
            }
        ])
        self.assertQuerySetEqual(
            DatasetElement.objects.filter(set__dataset_id=clone.id).values_list("set__name", "element__name").order_by("element__name", "set__name"),
            [
                ("dev", "Volume 1"),
                ("dev", "Volume 1, page 1r"),
                ("test", "Volume 1, page 1r"),
            ]
        )

    def test_clone_existing_name(self):
        self.corpus.datasets.create(name="Clone of First Dataset", creator=self.user)
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), [
            "First Dataset",
            "Clone of First Dataset",
            "Clone of First Dataset 1",
            "Second Dataset",
        ])

        data = response.json()
        data.pop("created")
        data.pop("updated")
        clone = Dataset.objects.get(id=data["id"])
        dev_clone, test_clone, train_clone = clone.sets.all().order_by("name")
        cloned_sets = data.pop("sets")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(clone.id),
                "name": "Clone of First Dataset 1",
                "description": self.dataset.description,
                "creator": self.user.display_name,
                "corpus_id": str(self.corpus.id),
                "set_elements": {str(k.name): 0 for k in self.dataset.sets.all()},
                "state": DatasetState.Open.value,
                "task_id": None,
                "unique_elements": True,
            },
        )
        self.assertCountEqual(cloned_sets, [
            {
                "name": "train",
                "id": str(train_clone.id)
            },
            {
                "name": "test",
                "id": str(test_clone.id)
            },
            {
                "name": "dev",
                "id": str(dev_clone.id)
            }
        ])

    def test_clone_name_too_long(self):
        dataset = self.corpus.datasets.create(name="A" * 99, creator=self.user)
        self.client.force_login(self.user)
        with self.assertNumQueries(12):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": dataset.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        clone = Dataset.objects.get(id=data.pop("id"))
        self.assertEqual(clone.name, "Clone of " + "A" * 91)
        self.assertEqual(data["name"], clone.name)

    # DatasetElementDestroy

    def test_destroy_dataset_element_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_dataset_element_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_destroy_dataset_element_requires_contributor(self, has_access_mock):
        self.client.force_login(self.read_user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        self.assertEqual(train_set.set_elements.count(), 1)
        self.assertEqual(dev_set.set_elements.count(), 1)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the dataset to perform this action."})
        self.dataset.refresh_from_db()
        self.assertEqual(train_set.set_elements.count(), 1)
        self.assertEqual(dev_set.set_elements.count(), 1)

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.read_user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_destroy_dataset_element_set_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set": ["This field is required."]})

    def test_destroy_dataset_element_requires_open_dataset(self):
        self.client.force_login(self.user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        self.dataset.state = DatasetState.Error
        self.dataset.save()
        self.assertEqual(train_set.set_elements.count(), 1)
        self.assertEqual(dev_set.set_elements.count(), 1)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"dataset": ["Elements can only be removed from open Datasets."]})
        self.dataset.refresh_from_db()
        self.assertEqual(train_set.set_elements.count(), 1)
        self.assertEqual(dev_set.set_elements.count(), 1)

    def test_destroy_dataset_element_dataset_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No DatasetElement matches the given query."})

    def test_destroy_dataset_element_set_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=match"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No DatasetElement matches the given query."})

    def test_destroy_dataset_element_element_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No DatasetElement matches the given query."})

    def test_destroy_dataset_element_element_not_in_dataset(self):
        train_set = self.dataset.sets.get(name="train")
        train_set.set_elements.create(element=self.page1, set="train")
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page2.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No DatasetElement matches the given query."})

    def test_destroy_dataset_element_wrong_set(self):
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page2.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "No DatasetElement matches the given query."})

    def test_destroy_dataset_element(self):
        self.client.force_login(self.user)
        dev_set, _, train_set = self.dataset.sets.all().order_by("name")
        train_set.set_elements.create(element=self.page1)
        dev_set.set_elements.create(element=self.page1)
        self.assertEqual(train_set.set_elements.count(), 1)
        self.assertEqual(dev_set.set_elements.count(), 1)
        with self.assertNumQueries(4):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.dataset.refresh_from_db()
        self.assertEqual(train_set.set_elements.count(), 0)
        self.assertEqual(dev_set.set_elements.count(), 1)

    def test_populate_dataset_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_populate_dataset_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_populate_dataset_private_corpus(self, filter_rights_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-populate", kwargs={"pk": self.private_dataset.pk}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.user, Corpus, Role.Guest.value))

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_populate_dataset_requires_contributor(self, has_access_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the dataset's corpus."})
        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.corpus, Role.Contributor.value, skip_public=False))

    def test_populate_dataset_invalid_parameters(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={
                    "parent_id": "a",
                    "recursive": "b",
                    "types": "c",
                    "count": "d",
                    "sets": "e",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "count": ["A valid integer is required."],
            "parent_id": ["Must be a valid UUID."],
            "recursive": ["Must be a valid boolean."],
            "sets": ['Expected a dictionary of items but got type "str".'],
            "types": ['Expected a list of items but got type "str".'],
        })

    def test_populate_dataset_unexisting_values(self):
        self.client.force_login(self.user)
        type = self.private_corpus.types.create(slug="A")
        elt = self.private_corpus.elements.create(name="a", type_id=type.id)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={
                    "parent_id": str(elt.id),
                    "types": ["page", "A", "AAAAAA"],
                    "sets": {"test": .8, "A": .15, "AAAAA": .05},
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "parent_id": ["This element does not exist in the corpus of the dataset."],
            "sets": ["Some sets does not exist in the dataset: ['A', 'AAAAA']."],
            "types": ["Some types does not exist in the corpus of the dataset: ['A', 'AAAAAA']."],
        })

    def test_populate_dataset_count_max_value(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={"recursive": True},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "count": ["This value exceeds the number of filtered elements (6)."]
        })

    def test_populate_dataset_wrong_sets_ratio(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={"recursive": True, "sets": {"test": 0.13, "train": 0.37}},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets": ["The sum of all ratios must be equal to 1."]
        })

    def test_populate_dataset_min_sets_ratio(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={"recursive": True, "count": 5, "sets": {"test": 0.01, "train": 0.99}},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets": ["Ratio must be greater than 0.2 to contain at least one element."]
        })

    def test_populate_dataset_duplicate_entries(self):
        """Raises a 400 in case an element of the selection is already part of a set"""
        DatasetElement.objects.create(
            element=self.page1,
            set=self.dataset.sets.get(name="train"),
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={"recursive": True, "count": 1, "sets": {"test": 1}},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "__all__": ["Some filtered elements are already part of a set on this dataset."]
        })

    def test_populate_dataset(self):
        DatasetElement.objects.create(
            element=self.vol,
            set=self.dataset.sets.get(name="train"),
        )
        self.client.force_login(self.user)
        with force_constraints_immediate(), self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={
                    "recursive": True,
                    "types": ["page", "word"],
                    "count": 11,
                    "sets": {"train": .2, "dev": .7, "test": .1}},
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json(), {})
        # 2 elements + volume DatasetElement
        self.assertEqual(DatasetElement.objects.filter(set__dataset=self.dataset, set__name="train").count(), 3)
        # 7 elements + the final one (largest sets populated first)
        self.assertEqual(DatasetElement.objects.filter(set__dataset=self.dataset, set__name="dev").count(), 8)
        # 1 elements
        self.assertEqual(DatasetElement.objects.filter(set__dataset=self.dataset, set__name="test").count(), 1)
        # Ensure no element is found in multiple sets at once
        self.assertEqual(
            DatasetElement.objects.filter(set__dataset=self.dataset).values("element_id").distinct("element_id").count(),
            12
        )

    def test_populate_dataset_direct_children(self):
        self.client.force_login(self.user)
        with force_constraints_immediate(), self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={
                    "parent_id": str(self.vol.id),
                    "recursive": False,
                    "types": ["page", "word"],
                    "count": 3,
                    "sets": {"train": 1}
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json(), {})
        self.assertQuerysetEqual(
            (
                DatasetElement.objects
                .filter(set__dataset=self.dataset)
                .order_by("element__name")
                .values_list("set__name", "element__name")
            ),
            (
                ("train", "Volume 1, page 1r"),
                ("train", "Volume 1, page 1v"),
                ("train", "Volume 1, page 2r"),
            )
        )

    def test_populate_dataset_recursive_children(self):
        self.client.force_login(self.user)
        with force_constraints_immediate(), self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:dataset-populate", kwargs={"pk": self.dataset.pk}),
                data={
                    "parent_id": str(self.vol.id),
                    "recursive": True,
                    "types": ["page", "word"],
                    "count": 12,
                    "sets": {"train": 1}
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertQuerysetEqual(
            (
                DatasetElement.objects
                .filter(set__dataset=self.dataset)
                .order_by("element__name")
                .values_list("set__name", "element__name")
            ),
            (
                ("train", "DATUM"),
                ("train", "DATUM"),
                ("train", "DATUM"),
                ("train", "PARIS"),
                ("train", "PARIS"),
                ("train", "PARIS"),
                ("train", "ROY"),
                ("train", "ROY"),
                ("train", "ROY"),
                ("train", "Volume 1, page 1r"),
                ("train", "Volume 1, page 1v"),
                ("train", "Volume 1, page 2r"),
            ),
        )
